defmodule InstaScrapex.JSON do
  @moduledoc """
    This module's functions parse the JSON endpoints of Instagram
    and return it as a map.
  """

  use HTTPoison.Base

  @doc """
    Use this to get info about accounts

    Wraps HTTPoison.get!, but takes a username instead of url,
    and returns a Map in the response[:body]

    ## Examples

        iex> InstaScrapex.JSON.get_account!("nasa")
        %HTTPoison.Response{body: %{}, ...}

  """
  def get_account!(username, headers \\ [], options \\ []) do
    username
    |> account_url
    |> get!(headers, options)
  end

  @doc false
  def account_url(username) when is_binary(username) do
    "https://www.instagram.com/#{username}/?__a=1"
  end

  @doc """
    Use this to get info about posts

    Wraps HTTPoison.get!, but takes a media_code instead of url,
    and returns a Map in the response[:body]

    ## Examples

        iex> InstaScrapex.JSON.get_media!("BKgPf5ZgP97")
        %HTTPoison.Response{body: %{}, ...}

  """
  def get_media!(media_code, headers \\ [], options \\ []) do
    media_code
    |> media_url
    |> get!(headers, options)
  end

  @doc false
  def media_url(media_code) when is_binary(media_code) do
    "https://www.instagram.com/p/#{media_code}/?__a=1"
  end

  @doc """
    Use this to get a list of posts from a user
    use :max_id in params to paginate and get older posts

    Wraps HTTPoison.get!, but takes a username instead of url,
    and returns a Map in the response[:body]

    ## Examples

        iex> InstaScrapex.JSON.get_account_media!("nasa")
        %HTTPoison.Response{body: %{}, ...}

        iex> InstaScrapex.JSON.get_account_media!("nasa", [], [params: [max_id: "12341234"]])
        %HTTPoison.Response{body: %{}, ...}

  """
  def get_account_media!(username, headers \\ [], options \\ []) do
    username
    |> account_media_url
    |> get!(headers, options)
  end

  @doc """
    Use this to get a list of a post's comments
    use :last_comment_id in post_data to paginate and get older posts

    Wraps HTTPoison.post!, but takes a %{code: "BKgPf5ZgP97"} instead of url,
    and returns a Map in the response[:body]

    ## Examples

        iex> InstaScrapex.JSON.get_media_comments!(%{code: "BKgPf5ZgP97"})
        %HTTPoison.Response{body: %{}, ...}
        #-- or --
        :no_comments

        iex> InstaScrapex.JSON.get_media_comments!(%{code: "BKgPf5ZgP97", last_comment_id: 12341234, per_page: 20})
        %HTTPoison.Response{body: %{}, ...}

  """
  def get_media_comments!(post_data = %{code: code}, headers \\ [], options \\ []) do
    last_comment_id =
      post_data[:last_comment_id] ||
      (get_media!(code).body["graphql"]["shortcode_media"]["edge_media_to_comment"]["edges"] |> List.last)["node"]["id"]

    per_page = post_data[:per_page] || 20

    case last_comment_id do
      nil ->
        :no_comments
      _  ->
        body = "q=ig_shortcode(#{code}){comments.before(#{last_comment_id},#{per_page}){count,nodes{id,created_at,text,user{id,profile_pic_url,username,follows{count},followed_by{count},biography,full_name,media{count},is_private,external_url,is_verified}},page_info}}"

        query_url()
        |> post!(body, headers ++ fake_headers(), options)
    end
  end

  @doc """
    Use this to get a list of posts from a user by the user_id

    Wraps HTTPoison.post!, but takes a %{user_id: 1234} instead of url,
    and returns a Map in the response[:body]

    ## Examples

        iex> InstaScrapex.JSON.get_media_by_user_id!(%{user_id: 1234})
        %HTTPoison.Response{body: %{}, ...}
  """
  def get_media_by_user_id!(post_data = %{user_id: user_id}, headers \\ [], options \\ []) do
    count = post_data[:count] || 16

    comments = "comments{count}"

    body = "q=ig_user(#{user_id}){media.first(#{count}){count,nodes{caption,code,#{comments},date,dimensions{height,width},display_src,id,is_video,likes{count},owner{id,username,full_name,profile_pic_url,biography},thumbnail_src,video_views,video_url},page_info}}"

    query_url()
    |> post!(body, headers ++ fake_headers(), options)
  end

  @doc """
    Use this to get a list of a hashtags posts

    Wraps HTTPoison.post!, but takes a %{hashtag: "my_hashtag"} instead of url,
    and returns a Map in the response[:body]

    ## Examples

        iex> InstaScrapex.JSON.get_media_by_hashtag!(%{hashtag: "my_hashtag"})
        %HTTPoison.Response{body: %{}, ...}

        iex> InstaScrapex.JSON.get_media_by_hashtag!(%{hashtag: "my_hashtag", end_cursor: "ABCDFEGHIJKL"})
        %HTTPoison.Response{body: %{}, ...}
  """
  def get_media_by_hashtag!(post_data = %{hashtag: hashtag}, headers \\ [], options \\ []) do
    count = post_data[:count] || 9
    end_cursor = post_data[:end_cursor]

    start = case end_cursor do
              nil -> "first(#{count})"
              _ -> "after(#{end_cursor},#{count})"
            end

    body = """
      q=ig_hashtag(#{hashtag}){
        media.#{start}{
          count,
          nodes{
            caption,
            code,
            comments{count},
            date,
            dimensions{height,width},
            display_src,
            id,
            is_video,
            likes{count},
            owner{
              id,
              username,
              full_name,
              followed_by{count},
              profile_pic_url,
              biography
            },
            thumbnail_src,
            video_views,
            video_url
          },
          page_info
        }
      }
    """ |> clean_string

    query_url()
    |> post!(body, headers ++ fake_headers(), options)
  end

  @doc """
    Use this to get a list of posts by location

    Wraps HTTPoison.get!, but takes a location_id instead of url,
    and returns a Map in the response[:body]

   You can get a location_id by using get_search!

    ## Examples

        iex> InstaScrapex.JSON.get_media_by_location!(location_id = 12341234)
        %HTTPoison.Response{body: %{}, ...}

        iex> InstaScrapex.JSON.get_media_by_location!(12341234, [], [after: last_post_id = 12345679234])
        %HTTPoison.Response{body: %{}, ...}
  """
  def get_media_by_location!(location_id, headers \\ [], options \\ []) do
    query_id = 17881432870018455
    count = 12

    base_params = [
      query_id: query_id,
      id: location_id,
      first: count,
    ]

    {_, options} = Keyword.get_and_update(options, :params, fn current_value ->
      new_value =
        case current_value do
          nil -> base_params
          params -> params ++ base_params
        end

      {current_value, new_value}
    end)

    "https://www.instagram.com/graphql/query/"
    |> get!(headers, options)
  end

  @doc """
    Search for hashtags, places, users

    Wraps HTTPoison.get, but takes a string instead of url,
    and returns a Map in the response[:body]

    ## Examples

        iex> InstaScrapex.JSON.get_search!("vienna")
        %HTTPoison.Response{body: %{}, ...}

        iex> InstaScrapex.JSON.get_search!("tokyo")
        %HTTPoison.Response{body: %{}, ...}
  """
  def get_search!(keyword, headers \\ [], options \\ []) do
    "https://www.instagram.com/web/search/topsearch/?context=blended&query=#{keyword}"
    |> get!(headers, options)
  end

  @doc false
  def account_media_url(username) when is_binary(username) do
   "https://www.instagram.com/#{username}/media/"
  end

  @doc false
  def process_response_body(body) do
    body |> Poison.decode!
  end

  @doc false
  def query_url, do: "https://www.instagram.com/query/"

  @doc false
  def fake_headers do
    random = random_string()

    [
      "content-type": "application/x-www-form-urlencoded",
      "X-CSRFToken": random,
      "Cookie":  "csrftoken=#{random};",
      "Referer": "https://www.instagram.com/",
      "x-instagram-ajax": "1",
      "x-requested-with": "XMLHttpRequest",
    ]
  end

  def random_string(length \\ 10) do
    :crypto.strong_rand_bytes(length) |> Base.url_encode64 |> binary_part(0, length) |> String.replace(~r/[_-]/, "a")
  end

  def clean_string(string) do
    string
    |> String.replace("\n", "")
    |> String.replace(" ", "")
  end
end
